<?php

namespace Drupal\deprecation_status\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\deprecation_status\DataSource;

class GraphController extends ControllerBase {

  public function graphPage($target_version = 9) {
    $build = ['chartjs' => [
      '#children' =>
        '<script src="https://cdn.jsdelivr.net/npm/moment@2.24.0/moment.js"></script>
        <script src="https://cdn.jsdelivr.net/npm/chart.js@2.8.0"></script>

        <script>
        const verticalLinePlugin = {
          getLinePosition: function (chart, pointIndex) {
              const meta = chart.getDatasetMeta(0); // first dataset is used to discover X coordinate of a point
              const data = meta.data;
              return data[pointIndex.index]._model.x;
          },
          renderVerticalLine: function (chartInstance, pointIndex) {
              const lineLeftOffset = this.getLinePosition(chartInstance, pointIndex);
              const scale = chartInstance.scales[\'y-axis-0\'];
              const context = chartInstance.chart.ctx;

              // render vertical line
              context.beginPath();
              context.strokeStyle = \'#48494b\';
              context.moveTo(lineLeftOffset, scale.top);
              context.lineTo(lineLeftOffset, scale.bottom - 20);
              context.setLineDash([3, 3]);
              context.stroke();

              // write label
              context.fillStyle = "#48494b";
              context.textAlign = \'center\';
              context.fillText(pointIndex.label, lineLeftOffset - 2, scale.bottom - 10);
          },

          afterDatasetsDraw: function (chart, easing) {
              if (chart.config.markerLines) {
                  chart.config.markerLines.forEach(pointIndex => this.renderVerticalLine(chart, pointIndex));
              }
          }
          };

          Chart.plugins.register(verticalLinePlugin);</script>',

    ]];

    if ($target_version == 9) {
      return $build + $this->graphProjectsAndErrors();
    }
    else {
      return $build + $this->graphProjectsAndErrors(10) + $this->graph10Deprecations();
    }
  }

  public function graphProjectsAndErrors($version = 9) {

    if ($version == 9) {
      $markers = "
      {label: 'Drupal 8.9', index: 4},
      {label: 'Drupal 9.0', index: 27},
      {label: 'Drupal 9.1', index: 44},
      {label: 'Drupal 9.2', index: 67}";
    }
    else {
      $markers = "{label: 'Drupal 9.2', index: 9}";
    }

    $breakdown_colors = [
      'Explicitly Drupal ' . $version . ' compatible' => "#8CBD3F",
      'Only info.yml/composer.json problem found' => "#FDC70F",
      'All rector covered' => "#1F72B2",
      'Some covered by rector' => "#42509B",
      'Needs manual review' => "#EC6225",
      'No results' => "#000000",
    ];
    $labels = "'Top 50', 'Top 100', 'Top200', 'Top 500', 'Top 1000', 'All'";

    // Read project segment data.
    $data = [];
    $file = fopen(DataSource::getFullPath('breakdown_total_segments.csv', $version), 'r');
    $i = 0;
    while ($line = fgetcsv($file, 0, ";")) {
      if ($i != 0) {
        $segment = array_shift($line);
        $sum = array_sum($line);
        $row = [];
        foreach($line as $item) {
          $row[] = round($item/($sum/100), 2);
        }
        $data[] = $row;
      }
      $i++;
    }
    fclose($file);

    $datasets = [];
    $i = 0;
    foreach($breakdown_colors as $label => $color) {
      $raw = '{label: "'. $label . '", data: [';
      foreach($data as $values) {
        $raw .= $values[$i] . ',';
      }
      $datasets[] = trim($raw, ',') . '], backgroundColor: "' . $color . '"}';
      $i++;
    }
    $datasets = join(',', $datasets);

    $build['breakdown_segments'] = [
      '#markup' => '<h2>Percentage of contributed projects by status in usage groups</h2><p>Larger groups include results for smaller usage groups.</p>',
      '#children' => <<<PROJECTSEND
  <canvas id="SegmentsStatus"></canvas>
  <script type="text/javascript">
  var ctx = document.getElementById('SegmentsStatus').getContext('2d');
  var chart = new Chart(ctx, {
    type: 'bar',
    data: {
      labels: [$labels],
      datasets: [$datasets]
    },
    options: {
      aspectRatio: window.innerWidth < 800 ? 1 : 2,
      scales: {
        xAxes: [{
          stacked: true
        }],
        yAxes: [{
          stacked: true,
          ticks: { max: 100 }
        }]
      }
    }
  });
</script>
PROJECTSEND
    ];

    // =========================================================================

    $project_graphs = [
      //'top200' => 'Top 200 drupal.org contributed projects by status over time',
      'total' => 'All drupal.org contributed projects by status over time',
    ];
    foreach ($project_graphs as $graph_key => $graph_label) {

      // Read historic project data.
      $projects = [];
      $legend = [];
      $file = fopen(DataSource::getFullPath('breakdown_' . $graph_key . '_projects.csv', $version), 'r');
      $i = 0;
      while ($line = fgetcsv($file, 0, ";")) {
        if ($i == 0) {
          $legend = $line;
          array_shift($legend);
        }
        else {
          $time = array_shift($line);
          foreach($legend as $index => $label) {
            $projects[$label][$time] = $line[$index];
          }
        }
        $i++;
      }
      fclose($file);

      $breakdown_colors = [
        'Explicitly Drupal ' . $version . ' compatible' => "#8CBD3F",
        'Only info.yml/composer.json problem found' => "#FDC70F",
        'All rector covered' => "#1F72B2",
        'Some covered by rector' => "#42509B",
        'Needs manual review' => "#EC6225",
        'No results' => "#000000",
      ];

      $datasets = [];
      foreach($projects as $label => $dataset) {
        $raw = '{label: "'. $label . '", data: [';
        foreach($dataset as $time => $value) {
          $raw .= '{x: new Date(' . $time . '*1000), y: ' . $value . '},';
        }
        $raw = trim($raw, ',') . ']';
        if (isset($breakdown_colors[$label])) {
          $raw .= ', borderColor: "' . $breakdown_colors[$label] . '", backgroundColor: "' . $breakdown_colors[$label] . '"';
        }
        $raw .= '}';
        $datasets[] = $raw;
      }
      $datasets = join(',', $datasets);

      $build[$graph_key . '_projects'] = [
        '#markup' => '<h2>' . $graph_label . '</h2><p>Projects have errors that can be <a href="https://www.drupal.org/project/rector">automatically fixed with drupal-rector</a>.</p>',
        '#children' => <<<PROJECTSEND
  <canvas id="{$graph_key}ProjectStatus"></canvas>
  <script type="text/javascript">
  var ctx = document.getElementById('{$graph_key}ProjectStatus').getContext('2d');
  var chart = new Chart(ctx, {
    type: 'line',
    data: {
      datasets: [$datasets]
    },
    options: {
      aspectRatio: window.innerWidth < 800 ? 1 : 2,
      scales: {
        xAxes: [{
          type: 'time',
          time: {
            unit: 'week',
            max: new Date('2021-12-31')
          }
        }],
        yAxes: [{
          stacked: true,
          ticks: {
            beginAtZero: true
          }
        }]
      }
    },
    markerLines: [$markers]
  });
</script>
PROJECTSEND
      ];
    }

    // =========================================================================

    $count_graphs = [
      //'top200' => 'Top 200 drupal.org contributed projects by error counts over time',
      'total' => 'All drupal.org contributed projects by error counts over time',
    ];
    foreach ($count_graphs as $graph_key => $graph_label) {

      // Read historic count data.
      $projects = [];
      $legend = [];
      $file = fopen(DataSource::getFullPath('breakdown_' . $graph_key . '_error_counts.csv', $version), 'r');
      $i = 0;
      while ($line = fgetcsv($file, 0, ";")) {
        if ($i == 0) {
          $legend = $line;
          array_shift($legend);
        }
        else {
          $time = array_shift($line);
          foreach($legend as $index => $label) {
            $projects[$label][$time] = $line[$index];
          }
        }
        $i++;
      }
      fclose($file);

      $breakdown_colors = [
        'Zero' => "#8CBD3F",
        'Less than 5' => "#FDC70F",
        'Less than 10' => "#1F72B2",
        '10 or more' => "#42509B",
        'No results' => "#000000",
      ];

      $datasets = [];
      foreach($projects as $label => $dataset) {
        $raw = '{label: "'. $label . '", data: [';
        foreach($dataset as $time => $value) {
          $raw .= '{x: new Date(' . $time . '*1000), y: ' . $value . '},';
        }
        $raw = trim($raw, ',') . ']';
        if (isset($breakdown_colors[$label])) {
          $raw .= ', borderColor: "' . $breakdown_colors[$label] . '", backgroundColor: "' . $breakdown_colors[$label] . '"';
        }
        $raw .= '}';
        $datasets[] = $raw;
      }
      $datasets = join(',', $datasets);

      $build[$graph_key . '_counts'] = [
        '#markup' => '<h2>' . $graph_label . '</h2><p>Number of projects by how many errors are in them.</p>',
        '#children' => <<<PROJECTSEND
  <canvas id="{$graph_key}CountStatus"></canvas>
  <script type="text/javascript">
  var ctx = document.getElementById('{$graph_key}CountStatus').getContext('2d');
  var chart = new Chart(ctx, {
    type: 'line',
    data: {
      datasets: [$datasets]
    },
    options: {
      aspectRatio: window.innerWidth < 800 ? 1 : 2,
      scales: {
        xAxes: [{
          type: 'time',
          time: {
            unit: 'week',
            max: new Date('2021-12-31')
          }
        }],
        yAxes: [{
          stacked: true,
          ticks: {
            beginAtZero: true
          }
        }]
      }
    },
    markerLines: [$markers]
  });
</script>
PROJECTSEND
      ];
    }

    // =========================================================================

    $type_graphs = [
      //'top200' => 'Top 200 drupal.org contributed projects\' errors by fixability over time',
      'total' => 'All drupal.org contributed projects\'s errors by fixability over time',
    ];
    foreach ($type_graphs as $graph_key => $graph_label) {

      $legend_labels = [
        'info' => 'info.yml and composer.json',
        'rector' => 'Covered by rector',
        'drupal' => 'Not covered by rector',
        'other' => 'Non-Drupal core deprecations',
        'parse' => 'Parsing errors',
      ];
      $legend_colors = [
        'info.yml and composer.json' => "#DBCDBC",
        'Covered by rector' => "#1F72B2",
        'Not covered by rector' => "#EC6225",
        'Non-Drupal core deprecations' => "#C31A7F",
        'Parsing errors' => "#F1E70D",
      ];
      if ($graph_key == 'top200') {
        $legend_labels['archive'] = 'Historic deprecation errors';
        $legend_colors['Historic deprecation errors'] = "#CDCDCD";
      }
      $legend = array_keys($legend_labels);
      $labels = array_values($legend_labels);

      // Read historic error data.
      $errors = [];
      $file = fopen(DataSource::getFullPath('breakdown_' . $graph_key . '_error_types.csv', $version), 'r');
      $i = 0;
      while ($line = fgetcsv($file, 0, ";")) {
        if ($i > 0) {
          $time = array_shift($line);
          foreach($legend as $index => $label) {
            $errors[$labels[$index]][$time] = $line[$index];
          }
        }
        $i++;
      }
      fclose($file);

      $datasets = [];
      foreach($errors as $label => $dataset) {
        $raw = '{label: "'. $label . '", data: [';
        foreach($dataset as $time => $value) {
          $raw .= '{x: new Date(' . $time . '*1000), y: ' . $value . '},';
        }
        $raw = trim($raw, ',') . ']';
        if (!empty($legend_colors[$label])) {
          $raw .= ', borderColor: "' . $legend_colors[$label] . '", backgroundColor: "' . $legend_colors[$label] . '"';
        }
        $raw .= '}';
        $datasets[] = $raw;
      }
      $datasets = join(',', $datasets);

      $build[$graph_key . '_errors'] = [
        '#markup' => '<h2>' . $graph_label . '</h2><p><a href="https://www.drupal.org/project/rector">Use drupal-rector</a> to fix rector-covered errors. <a href="https://www.drupal.org/project/rector">Help expand rector coverage</a> to make all our lives easier.</p>',
        '#children' => <<<ERRORSEND
  <canvas id="{$graph_key}ErrorStatus"></canvas>
  <script type="text/javascript">
  var ctx = document.getElementById('{$graph_key}ErrorStatus').getContext('2d');
  var chart = new Chart(ctx, {
    type: 'line',
    data: {
      datasets: [$datasets]
    },
    options: {
      aspectRatio: window.innerWidth < 800 ? 1 : 2,
      scales: {
        xAxes: [{
          type: 'time',
          time: {
            unit: 'week',
            max: new Date('2021-12-31')
          }
        }],
        yAxes: [{
          stacked: true,
          ticks: {
            beginAtZero: true
          }
        }]
      }
    },
    markerLines: [$markers]
  });
</script>
ERRORSEND
      ];
    }

    return $build;
  }

  public function graph10Deprecations() {
    // git checkout `git rev-list -1 --before="April 20 2021" 9.3.x`

    // $ grep -r "@deprecated" core --exclude-dir=node_modules | wc -l
    // $ grep -r "@deprecated" core --exclude-dir=node_modules | grep -o "11.0.0" | wc -l

    // $ grep -ro "E_USER_DEPRECATED" core --exclude-dir=node_modules | wc -l
    // $ grep -r "E_USER_DEPRECATED" core --exclude-dir=node_modules | grep -o "11.0.0" | wc -l

    // Read historic core deprecation count data.
    $at_deprecated = '';
    $e_user_deprecated = '';
    $file = fopen(DataSource::getFullPath('core_deprecation_count.csv', 10), 'r');
    $i = 0;
    while ($line = fgetcsv($file, 0, ";")) {
      // Skip the header row.
      if ($i > 0) {
        $at_deprecated .= '{x: new Date("' . $line[0] . '"), y: ' . $line[1] . '},';
        $e_user_deprecated .= '{x: new Date("' . $line[0] . '"), y: ' . $line[2] . '},';
      }
      $i++;
    }
    fclose($file);
    $at_deprecated = trim($at_deprecated, ',');
    $e_user_deprecated = trim($e_user_deprecated, ',');

    $build['core'] = [
      '#markup' => '<h2><a href="https://api.drupal.org/api/drupal/deprecated/9.2.x">Deprecated APIs</a> will be removed from the Drupal 10 branch when it opens.</h2><p>Tracking based on @deprecated and E_USER_DEPRECATED mentions in Drupal 9 core.</p>',
      '#children' => <<<COREEND
  <canvas id="coreDeprecated"></canvas>
  <script type="text/javascript">
  var ctx = document.getElementById('coreDeprecated').getContext('2d');
  var chart = new Chart(ctx, {
    type: 'line',
    data: {
      datasets: [{
        label: 'Drupal core @deprecated mentions',
        borderColor: "#FDC70F",
        borderWidth: 4,
        data: [$at_deprecated],
        cubicInterpolationMode: 'monotone',
        fill: false,
        pointRadius: 4,
        pointHoverRadius: 10,
        pointBackgroundColor: "#FDC70F",
      },
      {
        label: 'Drupal core E_USER_DEPRECATED mentions',
        borderColor: "#1F72B2",
        borderWidth: 4,
        data: [$e_user_deprecated],
        cubicInterpolationMode: 'monotone',
        fill: false,
        pointRadius: 4,
        pointHoverRadius: 10,
        pointBackgroundColor: "#1F72B2",
      }]
    },
    options: {
      scales: {
        xAxes: [{
          type: 'time',
          time: {
            unit: 'week',
            max: new Date('2021-12-31')
          }
        }],
        yAxes: [{
          ticks: {
            beginAtZero: true
          }
        }]
      }
    }
  });
</script>
COREEND
    ];

    return $build;
  }

}
